"""
    The app tracks the performance of various endpoints over time.
    To bind, use the following lines of code:
        import dashboard
        from flask import Flask
        ...
        app = Flask(__name__)
        ...
        dashboard.bind(app)

    The dashboard with the results that are collected can be found at:
        localhost:5000/dashboard
"""

import os
from flask import Blueprint
from flask_monitoringdashboard.config import Config
from unittest import TestLoader

config = Config()
user_app = None


# get current location of the project
def loc():
    return os.path.abspath(os.path.dirname(__file__)) + '/'


# define the blueprint
blueprint = Blueprint('dashboard', __name__, template_folder=loc() + 'templates')


def bind(app, blue_print=None):
    """
        Binding the app to this object should happen before importing the routing-
        methods below. Thus, the importing statement is part of this function.
        :param app: the app for which the performance has to be tracked
        :param blue_print: the blueprint that contains the endpoints to be monitored
    """
    assert app is not None
    global user_app, blueprint
    user_app = app

    if blue_print:
        import os
        import datetime
        from flask import request
        log_dir = os.getenv('DASHBOARD_LOG_DIR')

        @blue_print.after_request
        def after_request(response):
            if log_dir:
                t1 = str(datetime.datetime.now())
                log = open(log_dir + "endpoint_hits.log", "a")
                log.write("\"{}\",\"{}\"\n".format(t1, request.endpoint))
                log.close()
            return response

    # Add all route-functions to the blueprint
    import flask_monitoringdashboard.routings

    # Add wrappers to the endpoints that have to be monitored
    from flask_monitoringdashboard.measurement import init_measurement
    blueprint.before_app_first_request(init_measurement)

    # register the blueprint to the app
    app.register_blueprint(blueprint, url_prefix='/' + config.link)



"""
    This file can be executed for developing purposes. It is not used, when the flask_monitoring_dashboard is
    attached to an existing flask application.
"""

from flask import Flask, redirect, url_for
import flask_monitoringdashboard as dashboard
import os

user_app = Flask(__name__)
here = os.path.abspath(os.path.dirname(__file__))
dashboard.config.init_from(file=here + '/config.cfg')


def get_session_id():
    # implement here your own custom function
    return '12345'


dashboard.config.get_group_by = get_session_id
dashboard.bind(app=user_app)


@user_app.route('/')
def main():
    return redirect(url_for('dashboard.index'))


if __name__ == '__main__':
    user_app.run(debug=True, host='0.0.0.0')



"""
    Create a StackInfo-object and it will spawn a different thread.
    That thread will sleep for 'average' ms (reformatted to seconds).
    Then it will log the stacktrace of all active threads.
    Moreover, it logs cpu- and memory-info.
"""

import time
import traceback
from threading import Thread, enumerate
import sys
import psutil


class StackInfo(object):
    def __init__(self, average):
        self.average = average
        self.stacktrace = ''
        self.cpu_percent = ''
        self.memory = ''

        try:
            thread = Thread(target=log_stack_trace, args=(self,))
            thread.start()
        except Exception:
            print('Can\'t log traceback information')
            traceback.print_exc()


def log_stack_trace(stack_info):
    # average is in ms, sleep requires seconds
    time.sleep(stack_info.average / 1000.0)

    # iterate through every active thread and get the stack-trace
    stack_list = []
    try:
        for th in enumerate():
            try:
                f = open('stacktrace.log', 'w+')
                stack_list.extend(['', str(th)])
                traceback.print_stack(sys._current_frames()[th.ident], file=f)
                f.close()
                f = open('stacktrace.log', 'r')
                stack_list.extend(f.readlines())
            except Exception:
                print('Exception occurred:')
                traceback.print_exc()
    except Exception:
        print('Exception occurred: ')
        traceback.print_exc()

    # Set the values in the object
    stack_info.stacktrace = '<br />'.join(stack_list)
    stack_info.cpu_percent = str(psutil.cpu_percent(interval=None, percpu=True))
    stack_info.memory = str(psutil.virtual_memory())



"""
    This file contains all unit tests for the monitor-rules-table in the database. (Corresponding to the file:
    'dashboard/database/monitor-rules.py')
    See __init__.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, NAME, TIMES


class TestMonitorRule(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_get_monitor_rules(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.monitor_rules import get_monitor_rules
        from flask_monitoringdashboard import config
        result = get_monitor_rules()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].endpoint, NAME)
        self.assertTrue(result[0].monitor)
        self.assertEqual(result[0].version_added, config.version)
        self.assertEqual(result[0].last_accessed, TIMES[0])

    def test_get_monitor_names(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.monitor_rules import get_monitor_names
        result = get_monitor_names()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].endpoint, NAME)

    def test_reset_monitor_endpoints(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.monitor_rules import reset_monitor_endpoints
        self.test_get_monitor_rules()
        reset_monitor_endpoints()
        from flask_monitoringdashboard.database.monitor_rules import get_monitor_rules
        self.assertEqual(get_monitor_rules(), [])

    def test_get_monitor_data(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.monitor_rules import get_monitor_rules, get_monitor_data
        # since all monitor-rules in the test-database have the 'monitor'-variable set to True, the outcome of both
        # functions is equivalent
        result1 = get_monitor_data()
        result2 = get_monitor_rules()
        self.assertEqual(len(result1), len(result2))
        self.assertEqual(result1[0].endpoint, result2[0].endpoint)
        self.assertEqual(result1[0].last_accessed, result2[0].last_accessed)
        self.assertEqual(result1[0].monitor, result2[0].monitor)
        self.assertEqual(result1[0].time_added, result2[0].time_added)
        self.assertEqual(result1[0].version_added, result2[0].version_added)


"""
    This file contains all unit tests for the monitor-rules-table in the database. (Corresponding to the file:
    'dashboard/database/monitor-rules.py')
    See __init__.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, mean, \
    EXECUTION_TIMES, TIMES, NAME, GROUP_BY, IP


class TestMonitorRule(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_add_outlier(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.outlier import session_scope, Outlier
        with session_scope() as db_session:
            self.assertEqual(db_session.query(Outlier).all(), [])
        # TODO: Complete function

    def test_get_outliers(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.outlier import get_outliers
        self.assertEqual(get_outliers(NAME), [])


"""
    Main class for adding all route-functions to user_app.
    Setup requires only to import this file. All other imports are done in this file
"""
from flask_monitoringdashboard import blueprint, loc, user_app
from flask.helpers import send_from_directory
from flask import redirect, url_for

# Import more route-functions
import flask_monitoringdashboard.routings.login
import flask_monitoringdashboard.routings.setup
import flask_monitoringdashboard.routings.result
import flask_monitoringdashboard.routings.export_data
import flask_monitoringdashboard.routings.measurements

# Provide a secret-key for using WTF-forms
if user_app.secret_key is None:
    user_app.secret_key = 'my-secret-key'


# Rule for serving static files
@blueprint.route('/static/<path:filename>')
def static(filename):
    return send_from_directory(loc() + 'static', filename)


# All rules below are for viewing the dashboard-pages
@blueprint.route('/')
def index():
    return redirect(url_for('dashboard.overview'))



"""
Contains all functions that returns results of all tests
"""
from flask_monitoringdashboard.database import session_scope, Tests, TestRun
from sqlalchemy import func, desc


def get_tests():
    """ Return all existing tests. """
    with session_scope() as db_session:
        result = db_session.query(Tests).all()
        db_session.expunge_all()
        return result


def add_test(name):
    """ Add a newly found test to the database. """
    with session_scope() as db_session:
        db_session.add(Tests(name=name))


def add_or_update_test(name, last_run, succeeded):
    """ Updates values of a test. """
    with session_scope() as db_session:
        if db_session.query(Tests).filter(Tests.name == name).count():
            db_session.query(Tests).filter(Tests.name == name). \
                update({Tests.lastRun: last_run, Tests.succeeded: succeeded})
        else:
            db_session.add(Tests(name=name, lastRun=last_run, succeeded=succeeded))


def add_test_result(name, exec_time, time, version, suite, iter):
    """ Add a test result to the database. """
    with session_scope() as db_session:
        db_session.add(
            TestRun(name=name, execution_time=exec_time, time=time, version=version, suite=suite, run=iter))


def get_suite_nr():
    """ Retrieves the number of the next suite to run. """
    with session_scope() as db_session:
        result = db_session.query(func.max(TestRun.suite).label('nr')).one()
        if result.nr is None:
            next_nr = 1
        else:
            next_nr = result.nr + 1
        return next_nr


def get_results():
    """ Return all entries of measurements with their average. The results are grouped by their name. """
    with session_scope() as db_session:
        result = db_session.query(TestRun.name,
                                  func.count(TestRun.execution_time).label('count'),
                                  func.avg(TestRun.execution_time).label('average')
                                  ).group_by(TestRun.name).order_by(desc('count')).all()
        db_session.expunge_all()
        return result


def get_res_current(version):
    """ Return entries of measurements with their average from the current project version only. """
    with session_scope() as db_session:
        result = db_session.query(TestRun.name,
                                  func.count(TestRun.execution_time).label('count'),
                                  func.avg(TestRun.execution_time).label('average')) \
            .filter(TestRun.version == version) \
            .group_by(TestRun.name).order_by(desc('count')).all()
        db_session.expunge_all()
        return result


def get_line_results():
    with session_scope() as db_session:
        result = db_session.query(TestRun.version,
                                  func.avg(TestRun.execution_time).label('avg'),
                                  func.min(TestRun.execution_time).label('min'),
                                  func.max(TestRun.execution_time).label('max'),
                                  func.count(TestRun.execution_time).label('count')
                                  ).group_by(TestRun.version).order_by(desc('count')).all()
        db_session.expunge_all()
        return result


def get_suites():
    with session_scope() as db_session:
        result = db_session.query(TestRun.suite).group_by(TestRun.suite).all()
        db_session.expunge_all()
        return result


def get_measurements(suite):
    """Return all measurements for some Travis build. Used for creating a box plot. """
    with session_scope() as db_session:
        result = db_session.query(TestRun).filter(TestRun.suite == suite).all()
        db_session.expunge_all()
        return result


def get_test_measurements(name, suite):
    """Return all measurements for some test of some Travis build. Used for creating a box plot. """
    with session_scope() as db_session:
        result = db_session.query(TestRun).filter(TestRun.name == name, TestRun.suite == suite).all()
        db_session.expunge_all()
        return result



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, login


class TestMeasurement(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_overview(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/measurements/overview').status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/measurements/overview').status_code)

    def test_heatmap(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/measurements/heatmap').status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/measurements/heatmap').status_code)

    def test_page_number_of_requests_per_endpoint(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/measurements/requests').status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/measurements/requests').status_code)

    def test_page_boxplot_per_version(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/measurements/versions').status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/measurements/versions').status_code)

    def test_page_boxplot_per_endpoint(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/measurements/endpoints').status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/measurements/endpoints').status_code)




import setuptools


def desc():
    info = open('README.md').read()
    try:
        return info + '\n\n' + open('CHANGELOG.rst').read()
    except IOError:
        return info

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="Flask-MonitoringDashboard",
    version="1.10.3",
    packages=setuptools.find_packages(),
    include_package_data=True,
    platforms='Any',
    zip_safe=False,
    test_suite='flask_monitoringdashboard.test.start_testing',
    url='https://github.com/flask-dashboard/Flask-MonitoringDashboard',
    author="Patrick Vogel & Thijs Klooster",
    author_email="p.p.vogel@student.rug.nl",
    description="A dashboard for automatic monitoring of Flask web-services",
    long_description=desc(),
    install_requires=required,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'],
    project_urls={
        'Bug Reports': 'https://github.com/flask-dashboard/Flask-MonitoringDashboard/issues',
        'PyPi': 'https://pypi.org/project/Flask-MonitoringDashboard/',
        'Documentation': 'http://flask-monitoringdashboard.readthedocs.io/',
        'Source': 'https://github.com/flask-dashboard/Flask-MonitoringDashboard/',
    },
)



"""
Contains all functions that access any functionCall-object
"""

import datetime

from sqlalchemy import func, desc, text, asc

from flask_monitoringdashboard import config
from flask_monitoringdashboard.database import session_scope, FunctionCall


def get_reqs_endpoint_day():
    """ Retrieves the number of requests per endpoint per day. """
    with session_scope() as db_session:
        query = text("""select strftime('%Y-%m-%d', time) AS newTime,
                               count(endpoint) AS cnt,
                               endpoint
                        from functioncalls 
                        group by newTime, endpoint""")
        result = db_session.execute(query)
        data = result.fetchall()
        return data


def add_function_call(time, endpoint, ip):
    """ Add a measurement to the database. """
    with session_scope() as db_session:
        group_by = None
        if config.get_group_by:
            group_by = config.get_group_by()
        call = FunctionCall(endpoint=endpoint, execution_time=time, version=config.version,
                            time=datetime.datetime.now(), group_by=str(group_by), ip=ip)
        db_session.add(call)


def get_times():
    """ Return all entries of measurements with the average and total number of execution times. The results are 
    grouped by their endpoint. """
    with session_scope() as db_session:
        result = db_session.query(FunctionCall.endpoint,
                                  func.count(FunctionCall.execution_time).label('count'),
                                  func.avg(FunctionCall.execution_time).label('average')
                                  ).group_by(FunctionCall.endpoint).order_by(desc('count')).all()
        db_session.expunge_all()
        return result


def get_data_from(time_from):
    """
        Returns all data in the FunctionCall table, for the export data option.
        This function returns all data after the time_from date.
    """
    with session_scope() as db_session:
        result = db_session.query(FunctionCall).filter(FunctionCall.time >= time_from).all()
        db_session.expunge_all()
        return result


def get_data():
    """
    Equivalent function to get_data_from, but returns all data.
    :return: all data from the database in the Endpoint-table.
    """
    return get_data_from(datetime.date(1970, 1, 1))


def get_data_per_version(version):
    """ Returns all data in the FuctionCall table, grouped by their version. """
    with session_scope() as db_session:
        result = db_session.query(FunctionCall.execution_time, FunctionCall.version). \
            filter(FunctionCall.version == version).all()
        db_session.expunge_all()
        return result


def get_versions(end=None):
    with session_scope() as db_session:
        result = db_session.query(FunctionCall.version,
                                  func.min(FunctionCall.time).label('startedUsingOn')). \
            filter((FunctionCall.endpoint == end) | (end is None)).group_by(FunctionCall.version).order_by(
            asc('startedUsingOn')).all()
        db_session.expunge_all()
    return result


def get_data_per_endpoint(end):
    with session_scope() as db_session:
        result = db_session.query(FunctionCall.execution_time, FunctionCall.endpoint). \
            filter(FunctionCall.endpoint == end).all()
        db_session.expunge_all()
        return result


def get_endpoints():
    with session_scope() as db_session:
        result = db_session.query(FunctionCall.endpoint,
                                  func.count(FunctionCall.endpoint).label('cnt')). \
            group_by(FunctionCall.endpoint).order_by(asc('cnt')).all()
        db_session.expunge_all()
        return result
